#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>

#include <open-mx.h>
#include <CL/cl.h>

#include "ccl_common.h"


pthread_barrier_t _ccl_barrier;


void _cclGetPlatformIDs(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x01*/
	struct{
		char id;
		cl_uint num_entries;
	} *ccl_request=request;
	struct{
		cl_int result;
		cl_uint num_platforms;
		cl_platform_id platform[MAX_NUM_PLATFORMS];
	} ccl_reply={.num_platforms=0};

	ccl_reply.result = clGetPlatformIDs(min(MAX_NUM_PLATFORMS, ccl_request->num_entries), ccl_reply.platform, &ccl_reply.num_platforms);
	if((ccl_reply.num_platforms > MAX_NUM_PLATFORMS) && (ccl_request->num_entries > MAX_NUM_PLATFORMS))
		ccl_reply.result = CL_OUT_OF_HOST_MEMORY;

	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply) - sizeof(cl_platform_id) * (MAX_NUM_PLATFORMS - min(min(MAX_NUM_PLATFORMS, ccl_request->num_entries), ccl_reply.num_platforms)), addr, TAG_REPLY|subtag);
}


void _cclGetPlatformInfo(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x02*/
	struct{
		char id;
		cl_platform_id platform;
		cl_platform_info param_name;
		size_t param_value_size;
	} *ccl_request=request;
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply={.param_value_size_ret=0};

	if(ccl_request->param_name == CL_PLATFORM_HOSTNAME){
		gethostname(ccl_reply.param_value, min(MAX_PARAM_VALUE_SIZE, ccl_request->param_value_size));
		ccl_reply.param_value[MAX_PARAM_VALUE_SIZE-1]= '\0';
		ccl_reply.param_value_size_ret = strlen(ccl_reply.param_value) + 1;
	}else
		ccl_reply.result = clGetPlatformInfo(ccl_request->platform, ccl_request->param_name, min(MAX_PARAM_VALUE_SIZE, ccl_request->param_value_size), ccl_reply.param_value, &ccl_reply.param_value_size_ret);

	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply) - (MAX_PARAM_VALUE_SIZE - ccl_reply.param_value_size_ret), addr, TAG_REPLY|subtag);
}


void _cclGetDeviceIDs(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x03*/
	struct{
		char id;
		cl_platform_id platform;
		cl_device_type device_type;
		cl_uint num_entries;
	} *ccl_request=request;
	struct{
		cl_int result;
		cl_uint num_devices;
		cl_device_id device[MAX_NUM_DEVICES];
	} ccl_reply={.num_devices=0};

	ccl_reply.result = clGetDeviceIDs(ccl_request->platform, ccl_request->device_type, min(MAX_NUM_DEVICES, ccl_request->num_entries), ccl_reply.device, &ccl_reply.num_devices);
	if((ccl_reply.num_devices > MAX_NUM_DEVICES) && (ccl_request->num_entries > MAX_NUM_DEVICES))
		ccl_reply.result = CL_OUT_OF_HOST_MEMORY;

	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply) - sizeof(cl_device_id) * (MAX_NUM_DEVICES - min(min(MAX_NUM_DEVICES, ccl_request->num_entries), ccl_reply.num_devices)), addr, TAG_REPLY|subtag);
}


void _cclGetDeviceInfo(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x04*/
	struct{
		char id;
		cl_device_id device;
		cl_device_info param_name;
		size_t param_value_size;
	} *ccl_request=request;
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply={.param_value_size_ret=0};

	ccl_reply.result = clGetDeviceInfo(ccl_request->device, ccl_request->param_name, min(MAX_PARAM_VALUE_SIZE, ccl_request->param_value_size), ccl_reply.param_value, &ccl_reply.param_value_size_ret);

	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply) - (MAX_PARAM_VALUE_SIZE - ccl_reply.param_value_size_ret), addr, TAG_REPLY|subtag);
}


void _cclCreateContext(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x05*/
	struct{
		char id;
		cl_context_properties properties[MAX_NUM_PROPERTIES];
		cl_uint num_devices;
		cl_device_id devices[MAX_NUM_DEVICES];
	} *ccl_request=request;
	struct{
		cl_context result;
		cl_int errcode_ret;
	} ccl_reply;

	ccl_reply.result = clCreateContext(ccl_request->properties, ccl_request->num_devices, ccl_request->devices, NULL, NULL, &ccl_reply.errcode_ret);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclCreateCommandQueue(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x0A*/
	struct{
		char id;
		cl_context context;
		cl_device_id device;
		cl_command_queue_properties properties;
	} *ccl_request=request;
	struct{
		cl_command_queue result;
		cl_int errcode_ret;
	} ccl_reply;

	ccl_reply.result = clCreateCommandQueue(ccl_request->context, ccl_request->device, ccl_request->properties, &ccl_reply.errcode_ret);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclCreateBuffer(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x0E*/
	struct{
		char id;
		cl_context context;
		cl_mem_flags flags;
		size_t size;
	} *ccl_request=request;
	struct{
		cl_mem result;
		cl_int errcode_ret;
	} ccl_reply;

	ccl_reply.result = clCreateBuffer(ccl_request->context, ccl_request->flags, ccl_request->size, NULL, &ccl_reply.errcode_ret);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclEnqueueReadBuffer(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x10*/
	struct{
		char id;
		cl_command_queue command_queue;
		cl_mem buffer;
		size_t offset;
		size_t cb;
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} *ccl_request=request;
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;
	void *ptr;
	int b, cb2;

	ptr = malloc(ccl_request->cb);
	ccl_reply.result = clEnqueueReadBuffer(ccl_request->command_queue, ccl_request->buffer, CL_TRUE, ccl_request->offset, ccl_request->cb, ptr, ccl_request->num_events_in_wait_list, (ccl_request->num_events_in_wait_list == 0) ? NULL : ccl_request->event_wait_list, ccl_request->event_is_null ? NULL : &ccl_reply.event);
	
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
	if(ccl_reply.result == CL_SUCCESS)
		for(b=0; b<ccl_request->cb; b+=MAX_OMX_PACKET_SIZE){
			cb2 = min(MAX_OMX_PACKET_SIZE, ccl_request->cb - b);
			_ccl_omx_send_packet((void *)ptr+b, cb2, addr, TAG_DATA_PACKET|subtag);
		}

	free(ptr);
}


void _cclEnqueueWriteBuffer(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x11*/
	struct{
		char id;
		cl_command_queue command_queue;
		cl_mem buffer;
		size_t offset;
		size_t cb;
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} *ccl_request=request;
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;
	void *ptr;
	int b, cb2;

	ptr = malloc(ccl_request->cb);
	for(b=0; b<ccl_request->cb; b+=MAX_OMX_PACKET_SIZE){
		cb2 = min(MAX_OMX_PACKET_SIZE, ccl_request->cb - b);
		_ccl_omx_recv_packet((void *)ptr+b, cb2, TAG_DATA_PACKET|subtag, ALL_BITS, NULL);
	}

	ccl_reply.result = clEnqueueWriteBuffer(ccl_request->command_queue, ccl_request->buffer, CL_TRUE, ccl_request->offset, ccl_request->cb, ptr, ccl_request->num_events_in_wait_list, (ccl_request->num_events_in_wait_list == 0) ? NULL : ccl_request->event_wait_list, ccl_request->event_is_null ? NULL : &ccl_reply.event);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);

	free(ptr);
}


void _cclCreateProgramWithSource(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x29*/
	struct{
		char id;
		cl_context context;
		cl_uint count;
		size_t lengths[MAX_NUM_LENGTHS];
		char strings[MAX_STRINGS_SIZE];
	} *ccl_request=request;
	struct{
		cl_program result;
		cl_int errcode_ret;
	} ccl_reply;
	char **strings,*p=ccl_request->strings;
	int s;

	strings = malloc(sizeof(char *) * ccl_request->count);
	for(s=0; s<ccl_request->count; s++){
		(strings)[s] = (char *)malloc(sizeof(char) * ccl_request->lengths[s]);
		memcpy((strings)[s], p, ccl_request->lengths[s]);
		p += ccl_request->lengths[s];
	}

	ccl_reply.result = clCreateProgramWithSource(ccl_request->context, ccl_request->count, (const char **)strings, ccl_request->lengths, &ccl_reply.errcode_ret);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclBuildProgram(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x2D*/
	struct{
		char id;
		cl_program program;
		cl_uint num_devices;
		cl_device_id device_list[MAX_NUM_DEVICES];
		char options[MAX_OPTIONS_SIZE];
	} *ccl_request=request;
	struct{
		cl_int result;
	} ccl_reply;

	ccl_reply.result = clBuildProgram(ccl_request->program, ccl_request->num_devices, ccl_request->device_list, ccl_request->options, NULL, NULL);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclCreateKernel(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x31*/
	struct{
		char id;
		cl_program program;
		char kernel_name[MAX_KERNEL_NAME_SIZE];
	} *ccl_request=request;
	struct{
		cl_kernel result;
		cl_int errcode_ret;
	} ccl_reply;

	ccl_reply.result = clCreateKernel(ccl_request->program, ccl_request->kernel_name, &ccl_reply.errcode_ret);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclSetKernelArg(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x35*/
	struct{
		char id;
		cl_kernel kernel;
		cl_uint arg_index;
		size_t arg_size;
		char arg_value[MAX_ARG_VALUE_SIZE];
	} *ccl_request=request;
	struct{
		cl_int result;
	} ccl_reply;

	ccl_reply.result = clSetKernelArg(ccl_request->kernel, ccl_request->arg_index, ccl_request->arg_size, ccl_request->arg_value);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclEnqueueNDRangeKernel(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x38*/
	struct{
		char id;
		cl_command_queue command_queue;
		cl_kernel kernel;
		cl_uint work_dim;
		char global_work_offset_is_null;
		size_t global_work_offset[MAX_NUM_WORK_DIMS];
		size_t global_work_size[MAX_NUM_WORK_DIMS];
		char local_work_size_is_null;
		size_t local_work_size[MAX_NUM_WORK_DIMS];
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} *ccl_request=request;
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;

	ccl_reply.result = clEnqueueNDRangeKernel(ccl_request->command_queue, ccl_request->kernel, ccl_request->work_dim, ccl_request->global_work_offset_is_null ? NULL : ccl_request->global_work_offset, ccl_request->global_work_size, ccl_request->local_work_size_is_null ? NULL : ccl_request->local_work_size, ccl_request->num_events_in_wait_list, (ccl_request->num_events_in_wait_list == 0) ? NULL : ccl_request->event_wait_list, ccl_request->event_is_null ? NULL : &ccl_reply.event);
	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply), addr, TAG_REPLY|subtag);
}


void _cclGetEventProfilingInfo(void *request, omx_endpoint_addr_t addr, uint32_t subtag){/*0x48*/
	struct{
		char id;
		cl_event event;
		cl_profiling_info param_name;
		size_t param_value_size;
	} *ccl_request=request;
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply={.param_value_size_ret=0};

	ccl_reply.result = clGetEventProfilingInfo(ccl_request->event, ccl_request->param_name, min(MAX_PARAM_VALUE_SIZE, ccl_request->param_value_size), ccl_reply.param_value, &ccl_reply.param_value_size_ret);

	_ccl_omx_send_packet(&ccl_reply, sizeof(ccl_reply) - (MAX_PARAM_VALUE_SIZE - ccl_reply.param_value_size_ret), addr, TAG_REPLY|subtag);
}


void _connect(void *data, omx_endpoint_addr_t addr, uint32_t subtag){/*0x00*/
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id;
	omx_endpoint_addr_t remote_endpoint_addr;

	if(omx_decompose_endpoint_addr(addr, &remote_nic_id, &remote_endpoint_id) != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_decompose_endpoint_addr");
	if(omx_connect(_ccl_local_endpoint, remote_nic_id, remote_endpoint_id, FILTER_OPENCCL, TIMEOUT_CONNECT, &remote_endpoint_addr))
		_ccl_perror_and_exit("omx_connect");
	_ccl_omx_send_packet(NULL, 0, addr, TAG_REPLY|subtag);
}


void * dispatcher_thread_routine(void *arg){
	int thid;
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id;
	omx_status_t status;
	struct{
		char id;
		char data[MAX_REQ_PACKET_SIZE];
	} ccl_request;
	void (*handler[])(void *, omx_endpoint_addr_t, uint32_t) = {/*0x00*/_connect, /*0x01*/_cclGetPlatformIDs, /*0x02*/_cclGetPlatformInfo, /*0x03*/_cclGetDeviceIDs, /*0x04*/_cclGetDeviceInfo, /*0x05*/_cclCreateContext, /*0x06*/NULL, /*0x07*/NULL, /*0x08*/NULL, /*0x09*/NULL, /*0x0A*/_cclCreateCommandQueue, /*0x0B*/NULL, /*0x0C*/NULL, /*0x0D*/NULL, /*0x0E*/_cclCreateBuffer, /*0x0F*/NULL, /*0x10*/_cclEnqueueReadBuffer, /*0x11*/_cclEnqueueWriteBuffer, /*0x12*/NULL, /*0x13*/NULL, /*0x14*/NULL, /*0x15*/NULL, /*0x16*/NULL, /*0x17*/NULL, /*0x18*/NULL, /*0x19*/NULL, /*0x1A*/NULL, /*0x1B*/NULL, /*0x1C*/NULL, /*0x1D*/NULL, /*0x1E*/NULL, /*0x1F*/NULL, /*0x20*/NULL, /*0x21*/NULL, /*0x22*/NULL, /*0x23*/NULL, /*0x24*/NULL, /*0x25*/NULL, /*0x26*/NULL, /*0x27*/NULL, /*0x28*/NULL, /*0x29*/_cclCreateProgramWithSource, /*0x2A*/NULL, /*0x2B*/NULL, /*0x2C*/NULL, /*0x2D*/_cclBuildProgram, /*0x2E*/NULL, /*0x2F*/NULL, /*0x30*/NULL, /*0x31*/_cclCreateKernel, /*0x32*/NULL, /*0x33*/NULL, /*0x34*/NULL, /*0x35*/_cclSetKernelArg, /*0x36*/NULL, /*0x37*/NULL, /*0x38*/_cclEnqueueNDRangeKernel, /*0x39*/NULL, /*0x3A*/NULL, /*0x3B*/NULL, /*0x3C*/NULL, /*0x3D*/NULL, /*0x3E*/NULL, /*0x3F*/NULL, /*0x40*/NULL, /*0x41*/NULL, /*0x42*/NULL, /*0x43*/NULL, /*0x44*/NULL, /*0x45*/NULL, /*0x46*/NULL, /*0x47*/NULL, /*0x48*/_cclGetEventProfilingInfo};

	thid = (long)((int *)arg);

	pthread_barrier_wait(&_ccl_barrier);
	
	while(1){
		printf("%d: waiting...\n", thid);
		_ccl_omx_recv_packet(&ccl_request, sizeof(ccl_request), TAG_REQUEST, ALL_EXCEPT_SBTG, &status);
		if(omx_decompose_endpoint_addr(status.addr, &remote_nic_id, &remote_endpoint_id) != OMX_SUCCESS)
			_ccl_perror_and_exit("omx_decompose_endpoint_addr");
		printf("%d: %d bytes from 0x%016lX:%d [reqid=0x%02X, subtag=0x%016lX]\n", thid, status.msg_length, remote_nic_id, remote_endpoint_id, ccl_request.id, status.match_info);
		handler[(int)ccl_request.id](&ccl_request, status.addr, status.match_info);
	}

}


int main(int argc, char *argv[]){
	cl_int ret_code;
	cl_platform_id platform;
	cl_uint platforms;
	uint64_t local_nic_id;
	uint32_t local_endpoint_id;
	omx_endpoint_addr_t local_endpoint_addr;
	pthread_t dispatcher_thread[MAX_NUM_THREADS];
	int i;

	ret_code = clGetPlatformIDs(1, &platform, &platforms);
	if((ret_code != CL_SUCCESS) || (platforms == 0))
		_ccl_perror_and_exit("OpenCL");

	if(omx_init() != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_init");
	if(omx_open_endpoint(OMX_ANY_NIC, OMX_ANY_ENDPOINT, FILTER_OPENCCL, NULL, 0, &_ccl_local_endpoint) != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_open_endpoint");

	if(omx_get_endpoint_addr(_ccl_local_endpoint, &local_endpoint_addr) != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_get_endpoint_addr");
	if(omx_decompose_endpoint_addr(local_endpoint_addr, &local_nic_id, &local_endpoint_id) != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_decompose_endpoint_addr");

	printf("ccl_daemon is listening on nic 0x%016lX, endpoint 0x%08x\n", local_nic_id, local_endpoint_id);

	if(pthread_barrier_init(&_ccl_barrier, NULL, MAX_NUM_THREADS) != 0)
		_ccl_perror_and_exit("pthread_barrier_init");
	for(i=0; i<MAX_NUM_THREADS; i++)
		if(pthread_create(&dispatcher_thread[i], NULL, dispatcher_thread_routine, (int *)(long)i) != 0)
			_ccl_perror_and_exit("pthread_create");
	for(i=0; i<MAX_NUM_THREADS; i++)
		pthread_join(dispatcher_thread[i], NULL);
	exit(1);
}
