#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include <CL/cl.h>


#define MAX_FILENAME_SIZE 256
#define MAX_ANSWERS_SIZE 8192

#define min(a,b) (a<b?a:b)


char answers[MAX_ANSWERS_SIZE]="", *next_answer=NULL;


void read_answers(){
	FILE *file;
	char c, filename[MAX_FILENAME_SIZE];
	int i=0;

	printf("Reading answers\n");
	printf("\tfilename="); scanf("%s", filename);
	file = fopen(filename, "r");
	if(file != NULL){
		while(!feof(file)){
			fread(&c, 1, sizeof(char), file);
			if(!isspace(c))
				answers[i++] = c;
			else
				if(answers[i-1] != ' ')
					answers[i++] = ' ';
		}
		fclose(file);
	}else
		answers[0] = '\0';
	next_answer = answers;
}


void write_answers(){
	FILE *file;
	char filename[MAX_FILENAME_SIZE];

	answers[strlen(answers)-5] = '\0';
	printf("Writing answers\n");
	printf("\tfilename="); scanf("%s", filename);
	file = fopen(filename, "w");
	fwrite(answers, 1, sizeof(char) * strlen(answers), file);
	fclose(file);
}


void my_scanf0(char *format, void *ptr){
	#define MAX_ANSWER_SIZE 128
	char s[MAX_ANSWER_SIZE];

	if(next_answer == NULL)
		next_answer = answers;
	if(*next_answer != '\0'){
		sscanf(next_answer, "%s", s);
		printf("%s\n", s);
	}else{
		scanf("%s", s);
		strcat(answers, s);
		strcat(answers, " ");
	}
	sscanf(s, format, ptr);
	next_answer += (strlen(s) + 1);
}


void my_scanf1(char *msg, char *format, void *ptr){
	printf("%s", msg);
	my_scanf0(format, ptr);
}


void my_scanf2(char *msg, int msg_var, char *format, void *ptr){
	printf(msg, msg_var);
	my_scanf0(format, ptr);
}


/*char ** read_program_sources(int nfiles, ...){
	FILE *file;
	long size;
	va_list ap;
	int n;
	char **strings;

	strings = malloc(sizeof(char*) * nfiles);
	va_start(ap, nfiles);
	for(n=0; n<nfiles; n++){
		file = fopen(va_arg(ap, char*), "r");
		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);
		(strings)[n] = (char *)malloc(sizeof(char) * size + 1);
		fread((strings)[n], 1, sizeof(char) * size, file);
		(strings)[n][size] = '\0';
		fclose(file);
	}
	va_end(ap);

	return(strings);
}*/


char ** read_program_sources(int nfiles, char filename[][MAX_FILENAME_SIZE]){
	FILE *file;
	long size;
	int n;
	char **strings;

	strings = malloc(sizeof(char*) * nfiles);
	for(n=0; n<nfiles; n++){
		file = fopen(filename[n], "r");
		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);
		(strings)[n] = (char *)malloc(sizeof(char) * size + 1);
		fread((strings)[n], 1, sizeof(char) * size, file);
		(strings)[n][size] = '\0';
		fclose(file);
	}

	return(strings);
}


void test_cclGetPlatformIDs(){/*0x01*/
	#define MAX_PLATFORMS 16
	cl_uint num_platforms;
	cl_platform_id platform[MAX_PLATFORMS];
	cl_int result;
	int p;

	printf("Testing \"cclGetPlatformIDs\"\n");
	result = clGetPlatformIDs(MAX_PLATFORMS, platform, &num_platforms);
	printf("\tplatforms: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_platforms);
		for(p=0; p<min(num_platforms,MAX_PLATFORMS); p++)
			printf("\t\t%d: %#04lx\n", p, (long)platform[p]);
	}else
		printf("error!\n");
}


void test_cclGetPlatformInfo(){/*0x02*/
	#define MAX_PINFO_SIZE 256
	cl_platform_id platform;
	cl_int result;
	char platform_profile[MAX_PINFO_SIZE], platform_version[MAX_PINFO_SIZE], platform_name[MAX_PINFO_SIZE], platform_vendor[MAX_PINFO_SIZE], platform_extensions[MAX_PINFO_SIZE];
	size_t param_size;

	printf("Testing \"cclGetPlatformInfo\"\n");
	my_scanf1("\tplatform=", "%lx", (long *)&platform);
	result = clGetPlatformInfo(platform, CL_PLATFORM_PROFILE, sizeof(platform_profile), platform_profile, &param_size);
	printf("\t\tCL_PLATFORM_PROFILE: ");
	(result == CL_SUCCESS) ? printf("%s\n", platform_profile) : printf("error!\n");
	result = clGetPlatformInfo(platform, CL_PLATFORM_VERSION, sizeof(platform_version), platform_version, &param_size);
	printf("\t\tCL_PLATFORM_VERSION: ");
	(result == CL_SUCCESS) ? printf("%s\n", platform_version) : printf("error!\n");
	result = clGetPlatformInfo(platform, CL_PLATFORM_NAME, sizeof(platform_name), platform_name, &param_size);
	printf("\t\tCL_PLATFORM_NAME: ");
	(result == CL_SUCCESS) ? printf("%s\n", platform_name) : printf("error!\n");
	result = clGetPlatformInfo(platform, CL_PLATFORM_VENDOR, sizeof(platform_vendor), platform_vendor, &param_size);
	printf("\t\tCL_PLATFORM_VENDOR: ");
	(result == CL_SUCCESS) ? printf("%s\n", platform_vendor) : printf("error!\n");
	result = clGetPlatformInfo(platform, CL_PLATFORM_EXTENSIONS, sizeof(platform_extensions), platform_extensions, &param_size);
	printf("\t\tCL_PLATFORM_EXTENSIONS: ");
	(result == CL_SUCCESS) ? printf("%s\n", platform_extensions) : printf("error %d!\n", result);
}


void test_cclGetDeviceIDs(){/*0x03*/
	#define MAX_DEVICES 16
	cl_platform_id platform;
	cl_device_id device[MAX_DEVICES];
	cl_uint num_devices;
	cl_int result;
	int d;

	printf("Testing \"cclGetDeviceIDs\"\n");
	my_scanf1("\tplatform=", "%lx", (long *)&platform);
	result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, MAX_DEVICES, device, &num_devices);
	printf("\t\tCPU devices: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_devices);
		for(d=0; d<min(num_devices,MAX_DEVICES); d++)
			printf("\t\t\t%d: %#04lx\n", d, (long)device[d]);
	}else
		printf("error!\n");
	result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, MAX_DEVICES, device, &num_devices);
	printf("\t\tGPU devices: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_devices);
		for(d=0; d<min(num_devices,MAX_DEVICES); d++)
			printf("\t\t\t%d: %#04lx\n", d, (long)device[d]);
	}else
		printf("error!\n");
	result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_ACCELERATOR, MAX_DEVICES, device, &num_devices);
	printf("\t\tACCELERATOR devices: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_devices);
		for(d=0; d<min(num_devices,MAX_DEVICES); d++)
			printf("\t\t\t%d: %#04lx\n", d, (long)device[d]);
	}else
		printf("error!\n");
	result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_DEFAULT, MAX_DEVICES, device, &num_devices);
	printf("\t\tDEFAULT devices: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_devices);
		for(d=0; d<min(num_devices,MAX_DEVICES); d++)
			printf("\t\t\t%d: %#04lx\n", d, (long)device[d]);
	}else
		printf("error!\n");
	result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, MAX_DEVICES, device, &num_devices);
	printf("\t\tALL devices: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_devices);
		for(d=0; d<min(num_devices,MAX_DEVICES); d++)
			printf("\t\t\t%d: %#04lx\n", d, (long)device[d]);
	}else
		printf("error!\n");
}


void test_cclGetDeviceInfo(){/*0x04*/
	#define MAX_DINFO_SIZE 256
	cl_device_id device;
	cl_int result;
	cl_device_type device_type;
	cl_uint device_vendor_id, device_max_compute_units, device_max_work_item_dimensions, device_max_clock_frequency;
	char device_name[MAX_DINFO_SIZE], device_vendor[MAX_DINFO_SIZE], driver_version[MAX_DINFO_SIZE], device_opencl_c_version[MAX_DINFO_SIZE];
	size_t param_size;

	printf("Testing \"cclGetDeviceInfo\"\n");
	my_scanf1("\tdevice=", "%lx", (long *)&device);
	result = clGetDeviceInfo(device, CL_DEVICE_TYPE, sizeof(cl_device_type), &device_type, &param_size);
	printf("\t\tCL_DEVICE_TYPE: ");
	(result == CL_SUCCESS) ? printf("%#lx\n", device_type) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_VENDOR_ID, sizeof(cl_uint), &device_vendor_id, &param_size);
	printf("\t\tCL_DEVICE_VENDOR_ID: ");
	(result == CL_SUCCESS) ? printf("%#x\n", device_vendor_id) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), &device_max_compute_units, &param_size);
	printf("\t\tCL_DEVICE_MAX_COMPUTE_UNITS: ");
	(result == CL_SUCCESS) ? printf("%d\n", device_max_compute_units) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(cl_uint), &device_max_work_item_dimensions, &param_size);
	printf("\t\tCL_DEVICE_MAX_WORK_ITEM_DIMENSIONS: ");
	(result == CL_SUCCESS) ? printf("%d\n", device_max_work_item_dimensions) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(cl_uint), &device_max_clock_frequency, &param_size);
	printf("\t\tCL_DEVICE_MAX_CLOCK_FREQUENCY: ");
	(result == CL_SUCCESS) ? printf("%d\n", device_max_clock_frequency) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), &device_name, &param_size);
	printf("\t\tCL_DEVICE_NAME: ");
	(result == CL_SUCCESS) ? printf("%s\n", device_name) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(device_vendor), &device_vendor, &param_size);
	printf("\t\tCL_DEVICE_VENDOR: ");
	(result == CL_SUCCESS) ? printf("%s\n", device_vendor) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DRIVER_VERSION, sizeof(driver_version), &driver_version, &param_size);
	printf("\t\tCL_DRIVER_VERSION: ");
	(result == CL_SUCCESS) ? printf("%s\n", driver_version) : printf("error!\n");
	result = clGetDeviceInfo(device, CL_DEVICE_OPENCL_C_VERSION, sizeof(device_opencl_c_version), &device_opencl_c_version, &param_size);
	printf("\t\tCL_DEVICE_OPENCL_C_VERSION: ");
	(result == CL_SUCCESS) ? printf("%s\n", device_opencl_c_version) : printf("error!\n");
}


void CL_CALLBACK ctx_callback(const char *errinfo, const void *private_info, size_t cb, void *user_data){
	printf("\n********* ctx_callback *********\n");
	printf("\terrinfo=%s\n", errinfo);
	printf("\tprivate_info=%ldbytes\n", cb);
	printf("\tuser_data=%d\n", *((int *)user_data));
	printf("********* ctx_callback *********\n");
}


void test_cclCreateContext(){/*0x05*/
	#define MAX_PROPERTIES 8
	int p, d, num_properties, *callback_data;
	cl_uint num_devices;
	cl_context_properties properties[MAX_PROPERTIES];
	cl_device_id devices[MAX_DEVICES];
	cl_context ctx;
	cl_int errcode_ret;
	char callback;

	printf("Testing \"cclCreateContext\"\n");
	my_scanf1("\tnum_platforms=", "%d", &num_properties);
	for(p=0; p<num_properties*2; p+=2){
		properties[p] = CL_CONTEXT_PLATFORM;
		my_scanf2("\t\tplatform[%d]=", p/2, "%lx", &properties[p+1]);
	}
	properties[p] = 0;
	my_scanf1("\tnum_devices=", "%d", &num_devices);
	for(d=0; d<num_devices; d++)
		my_scanf2("\t\tdevice[%d]=", d, "%lx", (long *)&devices[d]);
	my_scanf1("\tcallback=", "%c", &callback);
	if(callback != 'n'){
		callback_data = (int *)malloc(sizeof(int));
		my_scanf1("\t\tcallback_data=", "%d", callback_data);
		ctx = clCreateContext(properties, num_devices, devices, ctx_callback, callback_data, &errcode_ret);
	}else
		ctx = clCreateContext(properties, num_devices, devices, NULL, NULL, &errcode_ret);
	printf("\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
}


void test_cclCreateContextFromType(){/*0x06*/
	#define MAX_PROPERTIES 8
	int p, num_properties, *callback_data;
	cl_context_properties properties[MAX_PROPERTIES];
	cl_context ctx;
	cl_int errcode_ret;
	char callback;

	printf("Testing \"cclCreateContextFromType\"\n");
	my_scanf1("\tnum_properties=", "%d", &num_properties);
	for(p=0; p<num_properties*2; p+=2){
		properties[p] = CL_CONTEXT_PLATFORM;
		my_scanf2("\t\tplatform[%d]=", p/2, "%lx", &properties[p+1]);
	}
	properties[p] = 0;
	printf("\tdevice_type=CL_DEVICE_TYPE_CPU\n");
	ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_CPU, NULL, NULL, &errcode_ret);
	printf("\t\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
	printf("\tdevice_type=CL_DEVICE_TYPE_GPU\n");
	ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_GPU, NULL, NULL, &errcode_ret);
	printf("\t\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
	printf("\tdevice_type=CL_DEVICE_TYPE_ACCELERATOR\n");
	ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_ACCELERATOR, NULL, NULL, &errcode_ret);
	printf("\t\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
	printf("\tdevice_type=CL_DEVICE_TYPE_DEFAULT\n");
	ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_DEFAULT, NULL, NULL, &errcode_ret);
	printf("\t\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
	printf("\tdevice_type=CL_DEVICE_TYPE_ALL\n");
	my_scanf1("\tcallback=", "%c", &callback);
	if(callback != 'n'){
		callback_data = (int *)malloc(sizeof(int));
		my_scanf1("\t\tcallback_data=", "%d", callback_data);
		ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_ALL, ctx_callback, callback_data, &errcode_ret);
	}else
		ctx = clCreateContextFromType(properties, CL_DEVICE_TYPE_ALL, NULL, NULL, &errcode_ret);
	printf("\t\tcontext: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error!\n");
}


void test_cclRetainContext(){/*0x07*/
	cl_context ctx;
	cl_int result;

	printf("Testing \"cclRetainContext\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	result = clRetainContext(ctx);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseContext(){/*0x08*/
	cl_context ctx;
	cl_int result;

	printf("Testing \"cclReleaseContext\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	result = clReleaseContext(ctx);
	printf("\trelease: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetContextInfo(){/*0x09*/
	#define MAX_PROPERTIES 8
	#define MAX_DEVICES 16
	cl_context ctx;
	cl_uint reference_count, num_devices;
	cl_int result;
	size_t param_size;
	cl_device_id device[MAX_DEVICES];
	cl_context_properties properties[MAX_PROPERTIES];
	int d, p;

	printf("Testing \"cclGetContextInfo\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	result = clGetContextInfo(ctx, CL_CONTEXT_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_CONTEXT_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
	result = clGetContextInfo(ctx, CL_CONTEXT_NUM_DEVICES, sizeof(cl_uint), &num_devices, &param_size);
	printf("\t\tCL_CONTEXT_NUM_DEVICES: ");
	(result == CL_SUCCESS) ? printf("%d\n", num_devices) : printf("error\n");
	printf("\t\tCL_CONTEXT_DEVICES: ");
	result = clGetContextInfo(ctx, CL_CONTEXT_DEVICES, sizeof(cl_device_id)*MAX_DEVICES, device, &param_size);
	if(result == CL_SUCCESS){
		for(d=0; d<num_devices; d++){
			if(d>0)
				printf("; ");
			printf("%#04lx", (long)device[d]);
		}
		printf("\n");
	}else
		printf("error\n");
	printf("\t\tCL_CONTEXT_PROPERTIES: ");
	result = clGetContextInfo(ctx, CL_CONTEXT_PROPERTIES, sizeof(cl_context_properties)*MAX_PROPERTIES, properties, &param_size);
	if(result == CL_SUCCESS){
		for(p=0; properties[p]!=0; p++){
			if(p>0)
				printf("; ");
			printf("%#04lx", (long)properties[p]);
		}
		printf("\n");
	}else
		printf("error\n");
}


void test_cclCreateCommandQueue(){/*0x0A*/
	int p, num_properties;
	cl_command_queue_properties cc_properties=0x00, cc_property;
	cl_device_id device;
	cl_command_queue command_queue;
	cl_context ctx;
	cl_int errcode_ret;

	printf("Testing \"cclCreateCommandQueue\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	my_scanf1("\tdevice=", "%lx", (long *)&device);
	printf("\t%#04x:OUT_OF_ORDER_EXEC_MODE_ENABLE; %#04x:PROFILING_ENABLE\n", CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, CL_QUEUE_PROFILING_ENABLE);
	my_scanf1("\tnum_properties=", "%d", &num_properties);
	for(p=0; p<num_properties; p++){
		my_scanf2("\t\tproperty[%d]=", p, "%lx", &cc_property);
		cc_properties |= cc_property;
	}
	command_queue = clCreateCommandQueue(ctx, device, cc_properties, &errcode_ret);
	printf("\tcommand_queue: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)command_queue) : printf("error!\n");
}


void test_cclRetainCommandQueue(){/*0x0B*/
	cl_command_queue command_queue;
	cl_int result;

	printf("Testing \"cclRetainCommandQueue\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	result = clRetainCommandQueue(command_queue);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseCommandQueue(){/*0x0C*/
	cl_command_queue command_queue;
	cl_int result;

	printf("Testing \"cclReleaseCommandQueue\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	result = clReleaseCommandQueue(command_queue);
	printf("\trelease: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetCommandQueueInfo(){/*0x0D*/
	cl_command_queue command_queue;
	cl_uint reference_count;
	cl_command_queue_properties properties;
	cl_int result;
	cl_context ctx;
	cl_device_id device;
	size_t param_size;

	printf("Testing \"cclGetCommandQueueInfo\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	result = clGetCommandQueueInfo(command_queue, CL_QUEUE_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_QUEUE_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
	result = clGetCommandQueueInfo(command_queue, CL_QUEUE_CONTEXT, sizeof(cl_context), &ctx, &param_size);
	printf("\t\tCL_QUEUE_CONTEXT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error\n");
	result = clGetCommandQueueInfo(command_queue, CL_QUEUE_DEVICE, sizeof(cl_device_id), &device, &param_size);
	printf("\t\tCL_QUEUE_DEVICE: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)device) : printf("error\n");
	result = clGetCommandQueueInfo(command_queue, CL_QUEUE_PROPERTIES, sizeof(cl_command_queue_properties), &properties, &param_size);
	printf("\t\tCL_QUEUE_PROPERTIES: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", properties) : printf("error\n");
}


void test_cclCreateBuffer(){/*0x0E*/
	int f, num_flags;
	cl_mem buffer;
	cl_context ctx;
	cl_mem_flags flag, flags=0x00;
	size_t buffer_size;
	cl_int errcode_ret;

	printf("Testing \"cclCreateBuffer\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	printf("\t%#04x:READ_WRITE; %#04x:WRITE_ONLY; %#04x:READ_ONLY; %#04x:USE_HOST_PTR; %#04x:ALLOC_HOST_PTR; %#04x:COPY_HOST_PTR\n", CL_MEM_READ_WRITE, CL_MEM_WRITE_ONLY, CL_MEM_READ_ONLY, CL_MEM_USE_HOST_PTR, CL_MEM_ALLOC_HOST_PTR, CL_MEM_COPY_HOST_PTR);
	my_scanf1("\tnum_flags=", "%d", &num_flags);
	for(f=0; f<num_flags; f++){
		my_scanf2("\t\tflag[%d]=", f, "%lx", &flag);
		flags |= flag;
	}
	my_scanf1("\tsize=", "%ld", &buffer_size);
	buffer = clCreateBuffer(ctx, flags, buffer_size, NULL, &errcode_ret);
	printf("\tbuffer: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)buffer) : printf("error!\n");
}


void test_cclCreateSubBuffer(){/*0x0F*/
	int f, num_flags;
	cl_mem buffer, subbuffer;
	cl_context ctx;
	cl_mem_flags flag, flags=0x00;
	cl_buffer_region buffer_create_info;
	cl_int errcode_ret;

	printf("Testing \"cclCreateSubBuffer\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	my_scanf1("\tbuffer=", "%lx", (long *)&buffer);
	printf("\t%#04x:READ_WRITE; %#04x:WRITE_ONLY; %#04x:READ_ONLY; %#04x:USE_HOST_PTR; %#04x:ALLOC_HOST_PTR; %#04x:COPY_HOST_PTR\n", CL_MEM_READ_WRITE, CL_MEM_WRITE_ONLY, CL_MEM_READ_ONLY, CL_MEM_USE_HOST_PTR, CL_MEM_ALLOC_HOST_PTR, CL_MEM_COPY_HOST_PTR);
	my_scanf1("\tnum_flags=", "%d", &num_flags);
	for(f=0; f<num_flags; f++){
		my_scanf2("\t\tflag[%d]=", f, "%lx", &flag);
		flags |= flag;
	}
	my_scanf1("\torigin=", "%ld", &buffer_create_info.origin);
	my_scanf1("\tsize=", "%ld", &buffer_create_info.size);
	subbuffer = clCreateSubBuffer(buffer, flags, CL_BUFFER_CREATE_TYPE_REGION, &buffer_create_info, &errcode_ret);
	printf("\tbuffer: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)subbuffer) : printf("error!\n");
}


void test_cclEnqueueReadBuffer(){/*0x10*/
	#define MAX_EVENTS 8
	cl_command_queue command_queue;
	cl_mem buffer;
	cl_int result;
	cl_event event_wait_list[MAX_EVENTS], event, *_event;
	cl_uint num_events_in_wait_list;
	cl_bool blocking_read;
	size_t offset, cb;
	void *ptr;
	char value_type, return_event, blocking;
	int i, e;

	printf("Testing \"cclEnqueueReadBuffer\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	my_scanf1("\tbuffer=", "%lx", (long *)&buffer);
	my_scanf1("\tblocking_read=", "%c", &blocking);
	blocking_read = (blocking == 'n') ? CL_FALSE : CL_TRUE;
	my_scanf1("\toffset=", "%ld", &offset);
	my_scanf1("\tcb=", "%ld", &cb);
	ptr = malloc(cb);
	my_scanf1("\tvalue_type=", "%c", &value_type);
	my_scanf1("\tnum_events_in_wait_list=", "%d", &num_events_in_wait_list);
	for(e=0; e<num_events_in_wait_list; e++)
		my_scanf2("\t\tevent[%d]=", e, "%lx", &event_wait_list[e]);
	my_scanf1("\tevent=", "%c", &return_event);
	_event = (return_event == 'n') ? NULL : &event;
	result = clEnqueueReadBuffer(command_queue, buffer, blocking_read, offset, cb, ptr, num_events_in_wait_list, event_wait_list, _event);
	printf("\tenqueue: ");
	if(result == CL_SUCCESS){
		switch(value_type){
			default:
			case 'i':
				for(i=0; i<cb/sizeof(int); i++){
					if(i > 0)
						printf(";");
					printf("%d", ((int *)ptr)[i]);
				}
				if(blocking == 'n'){
					sleep(1);
					printf("///");
					for(i=0; i<cb/sizeof(int); i++){
						if(i > 0)
							printf(";");
						printf("%d", ((int *)ptr)[i]);
					}
				}
				break;
		}
		printf("\n");
		if(return_event != 'n')
			printf("\tevent: %#04lx\n", (long)event);
	}else
		printf("error!\n");
	free(ptr);
}


void test_cclEnqueueWriteBuffer(){/*0x11*/
	cl_command_queue command_queue;
	cl_mem buffer;
	cl_int result;
	size_t offset, cb;
	char value_type;
	void *ptr;
	int i, int_value;

	printf("Testing \"cclEnqueueWriteBuffer\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	my_scanf1("\tbuffer=", "%lx", (long *)&buffer);
	my_scanf1("\toffset=", "%ld", &offset);
	my_scanf1("\tcb=", "%ld", &cb);
	ptr = malloc(cb);
	my_scanf1("\tvalue_type=", "%c", &value_type);
	switch(value_type){
		default:
		case 'i':
			my_scanf1("\t\tint_value=", "%d", &int_value);
			for(i=0; i<cb/sizeof(int); i++)
				((int *)ptr)[i] = int_value;
			break;
	}
	result = clEnqueueWriteBuffer(command_queue, buffer, CL_TRUE, offset, cb, ptr, 0, NULL, NULL);
	printf("\tenqueue: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
	free(ptr);
}


void test_cclRetainMemObject(){/*0x21*/
	cl_mem memobj;
	cl_int result;

	printf("Testing \"cclRetainMemObject\"\n");
	my_scanf1("\tmemobj=", "%lx", (long *)&memobj);
	result = clRetainMemObject(memobj);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseMemObject(){/*0x22*/
	cl_mem memobj;
	cl_int result;

	printf("Testing \"cclReleaseMemObject\"\n");
	my_scanf1("\tmemobj=", "%lx", (long *)&memobj);
	result = clReleaseMemObject(memobj);
	printf("\trelease: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetMemObjectInfo(){/*0x25*/
	cl_mem memobj, memobj2;
	cl_mem_object_type object_type;
	cl_mem_flags flags;
	void *host_ptr;
	cl_uint map_count, reference_count, result;
	cl_context ctx;
	size_t param_size, size, offset;

	printf("Testing \"cclGetMemObjectInfo\"\n");
	my_scanf1("\tmemobj=", "%lx", (long *)&memobj);
	result = clGetMemObjectInfo(memobj, CL_MEM_TYPE, sizeof(cl_mem_object_type), &object_type, &param_size);
	printf("\t\tCL_MEM_TYPE: ");
	(result == CL_SUCCESS) ? printf("%#04x\n", object_type) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_CONTEXT, sizeof(cl_context), &ctx, &param_size);
	printf("\t\tCL_MEM_CONTEXT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_ASSOCIATED_MEMOBJECT, sizeof(cl_mem), &memobj2, &param_size);
	printf("\t\tCL_MEM_ASSOCIATED_MEMOBJECT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)memobj2) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_FLAGS, sizeof(cl_mem_flags), &flags, &param_size);
	printf("\t\tCL_MEM_FLAGS: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", flags) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_SIZE, sizeof(size_t), &size, &param_size);
	printf("\t\tCL_MEM_SIZE: ");
	(result == CL_SUCCESS) ? printf("%ld\n", size) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_HOST_PTR, sizeof(void *), &host_ptr, &param_size);
	printf("\t\tCL_MEM_HOST_PTR: ");
	(result == CL_SUCCESS) ? printf("%p\n", host_ptr) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_MAP_COUNT, sizeof(cl_uint), &map_count, &param_size);
	printf("\t\tCL_MEM_MAP_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", map_count) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_MEM_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
	result = clGetMemObjectInfo(memobj, CL_MEM_OFFSET, sizeof(size_t), &offset, &param_size);
	printf("\t\tCL_MEM_OFFSET: ");
	(result == CL_SUCCESS) ? printf("%ld\n", offset) : printf("error\n");
}


void test_cclCreateProgramWithSource(){/*0x29*/
	#define MAX_FILENAMES 8
	cl_context ctx;
	cl_program program;
	cl_int errcode_ret;
	char filename[MAX_FILENAMES][MAX_FILENAME_SIZE], **strings;
	int f, num_filenames;

	printf("Testing \"cclCreateProgramWithSource\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	my_scanf1("\tnum_filenames=", "%d", &num_filenames);
	for(f=0; f<num_filenames; f++)
		my_scanf2("\t\tfilename[%d]=", f, "%s", filename[f]);
	strings = read_program_sources(num_filenames, filename);
	program = clCreateProgramWithSource(ctx, num_filenames, (const char **)strings, NULL, &errcode_ret);
	printf("\tprogram: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)program) : printf("error!\n");
}


void test_cclRetainProgram(){/*0x2B*/
	cl_program program;
	cl_int result;

	printf("Testing \"cclRetainProgram\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	result = clRetainProgram(program);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseProgram(){/*0x2C*/
	cl_program program;
	cl_int result;

	printf("Testing \"cclReleaseProgram\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	result = clReleaseProgram(program);
	printf("\trelease: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void CL_CALLBACK build_callback(cl_program program, void *user_data){
	printf("\n********* build_callback *********\n");
	printf("\tprogram=%p\n", program);
	printf("\tuser_data=%d\n", *((int *)user_data));
	printf("********* build_callback *********\n");
}


void test_cclBuildProgram(){/*0x2D*/
	cl_device_id devices[MAX_DEVICES];
	cl_program program;
	cl_uint num_devices;
	cl_int result;
	int d, *callback_data;
	char callback;

	printf("Testing \"cclBuildProgram\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	my_scanf1("\tnum_devices=", "%d", &num_devices);
	for(d=0; d<num_devices; d++)
		my_scanf2("\t\tdevice[%d]=", d, "%lx", (long *)&devices[d]);
	my_scanf1("\tcallback=", "%c", &callback);
	if(callback != 'n'){
		callback_data = (int *)malloc(sizeof(int));
		my_scanf1("\t\tcallback_data=", "%d", callback_data);
		result = clBuildProgram(program, num_devices, devices, "", build_callback, callback_data);
	}else
		result = clBuildProgram(program, num_devices, devices, "", NULL, NULL);
	printf("\tbuild: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetProgramInfo(){/*0x2F*/
	#define MAX_SOURCE_SIZE 2048
	#define MAX_DEVICES 16
	cl_program program;
	cl_uint reference_count, num_devices, result;
	cl_context ctx;
	cl_device_id device[MAX_DEVICES];
	size_t param_size, sizes[MAX_DEVICES];
	char source[MAX_SOURCE_SIZE];
	int s, d;

	printf("Testing \"cclGetProgramInfo\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	result = clGetProgramInfo(program, CL_PROGRAM_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_PROGRAM_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
	result = clGetProgramInfo(program, CL_PROGRAM_CONTEXT, sizeof(cl_context), &ctx, &param_size);
	printf("\t\tCL_PROGRAM_CONTEXT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error\n");
	result = clGetProgramInfo(program, CL_PROGRAM_NUM_DEVICES, sizeof(cl_uint), &num_devices, &param_size);
	printf("\t\tCL_PROGRAM_NUM_DEVICES: ");
	(result == CL_SUCCESS) ? printf("%d\n", num_devices) : printf("error\n");
	printf("\t\tCL_CONTEXT_DEVICES: ");
	result = clGetProgramInfo(program, CL_PROGRAM_DEVICES, sizeof(cl_device_id)*MAX_DEVICES, device, &param_size);
	if(result == CL_SUCCESS){
		for(d=0; d<num_devices; d++){
			if(d>0)
				printf("; ");
			printf("%#04lx", (long)device[d]);
		}
		printf("\n");
	}else
		printf("error\n");
	result = clGetProgramInfo(program, CL_PROGRAM_SOURCE, MAX_SOURCE_SIZE, source, &param_size);
	printf("\t\tCL_PROGRAM_SOURCE: ");
	(result == CL_SUCCESS) ? printf("%ld bytes\n", param_size) : printf("error\n");
	result = clGetProgramInfo(program, CL_PROGRAM_BINARY_SIZES, MAX_DEVICES * sizeof(size_t), &sizes, &param_size);
	printf("\t\tCL_PROGRAM_BINARY_SIZES: ");
	if(result == CL_SUCCESS){
		for(s=0; s<param_size/sizeof(size_t); s++){
			if(s>0)
				printf("; ");
			printf("%ld", sizes[s]);
		}
		printf("\n");
	}else
		printf("error\n");
}


void test_cclGetProgramBuildInfo(){/*0x30*/
	#define MAX_BUILD_OPTIONS_SIZE 128
	#define MAX_BUILD_LOG_SIZE 128
	cl_program program;
	cl_device_id device;
	cl_build_status status;
	cl_uint result;
	size_t param_size;
	char build_options[MAX_BUILD_OPTIONS_SIZE], build_log[MAX_BUILD_LOG_SIZE];

	printf("Testing \"cclGetProgramBuildInfo\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	my_scanf1("\tdevice=", "%lx", (long *)&device);
	result = clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_STATUS, sizeof(cl_build_status), &status, &param_size);
	printf("\t\tCL_PROGRAM_BUILD_STATUS: ");
	(result == CL_SUCCESS) ? printf("%#04x\n", status) : printf("error\n");
	result = clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_OPTIONS, MAX_BUILD_OPTIONS_SIZE, build_options, &param_size);
	printf("\t\tCL_PROGRAM_BUILD_OPTIONS: ");
	(result == CL_SUCCESS) ? printf("\"%s\"\n", build_options) : printf("error\n");
	result = clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, MAX_BUILD_LOG_SIZE, build_log, &param_size);
	printf("\t\tCL_PROGRAM_BUILD_LOG: ");
	(result == CL_SUCCESS) ? printf("\"%s\"\n", build_log) : printf("error\n");
}


void test_cclCreateKernel(){/*0x31*/
	#define MAX_KERNELNAME_SIZE 256
	cl_program program;
	cl_kernel kernel;
	cl_int errcode_ret;
	char kernel_name[MAX_KERNELNAME_SIZE];

	printf("Testing \"cclCreateKernel\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	my_scanf1("\tkernel_name=", "%s", kernel_name);
	kernel = clCreateKernel(program, kernel_name, &errcode_ret);
	printf("\tkernel: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)kernel) : printf("error!\n");
}


void test_cclCreateKernelsInProgram(){/*0x32*/
	#define MAX_KERNELS 8
	cl_program program;
	cl_kernel kernels[MAX_KERNELS];
	cl_uint num_kernels_ret, result;
	int k;

	printf("Testing \"cclCreateKernelsInProgram\"\n");
	my_scanf1("\tprogram=", "%lx", (long *)&program);
	result = clCreateKernelsInProgram(program, MAX_KERNELS, kernels, &num_kernels_ret);
	printf("\tkernels: ");
	if(result == CL_SUCCESS){
		printf("%d\n", num_kernels_ret);
		for(k=0; k<num_kernels_ret; k++)
			printf("\t\t%d: %#04lx\n", k, (long)kernels[k]);
	}else
		printf("error!\n");
}


void test_cclRetainKernel(){/*0x33*/
	cl_kernel kernel;
	cl_int result;

	printf("Testing \"cclRetainKernel\"\n");
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	result = clRetainKernel(kernel);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseKernel(){/*0x34*/
	cl_kernel kernel;
	cl_int result;

	printf("Testing \"cclReleaseKernel\"\n");
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	result = clReleaseKernel(kernel);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclSetKernelArg(){/*0x35*/
	cl_kernel kernel;
	cl_int result;
	cl_uint arg_index;
	cl_mem buffer;
	int int_value;
	size_t arg_size;
	void *arg_value_ptr;
	char arg_value_type;

	printf("Testing \"cclSetKernelArg\"\n");
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	my_scanf1("\targ_index=", "%d", &arg_index);
	my_scanf1("\targ_value_type=", "%c", &arg_value_type);
	switch(arg_value_type){
		default:
		case 'b':
			my_scanf1("\t\tbuffer=", "%lx", (long *)&buffer);
			arg_size = sizeof(cl_mem);
			arg_value_ptr = &buffer;
			break;
		case 'i':
			my_scanf1("\t\tint_value=", "%d", &int_value);
			arg_size = sizeof(int);
			arg_value_ptr = &int_value;
			break;
	}
	result = clSetKernelArg(kernel, arg_index, arg_size, arg_value_ptr);
	printf("\tset: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetKernelInfo(){/*0x36*/
	#define MAX_KERNEL_FUNCTION_NAME_SIZE 256
	cl_kernel kernel;
	cl_uint result, num_args, reference_count;
	cl_context ctx;
	cl_program program;
	size_t param_size;
	char kernel_name[MAX_KERNEL_FUNCTION_NAME_SIZE];

	printf("Testing \"cclGetKernelInfo\"\n");
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	result = clGetKernelInfo(kernel, CL_KERNEL_CONTEXT, sizeof(cl_context), &ctx, &param_size);
	printf("\t\tCL_KERNEL_CONTEXT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error\n");
	result = clGetKernelInfo(kernel, CL_KERNEL_PROGRAM, sizeof(cl_program), &program, &param_size);
	printf("\t\tCL_KERNEL_PROGRAM: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)program) : printf("error\n");
	result = clGetKernelInfo(kernel, CL_KERNEL_FUNCTION_NAME, MAX_KERNEL_FUNCTION_NAME_SIZE, kernel_name, &param_size);
	printf("\t\tCL_KERNEL_FUNCTION_NAME: ");
	(result == CL_SUCCESS) ? printf("\"%s\"\n", kernel_name) : printf("error\n");
	result = clGetKernelInfo(kernel, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &num_args, &param_size);
	printf("\t\tCL_KERNEL_NUM_ARGS: ");
	(result == CL_SUCCESS) ? printf("%d\n", num_args) : printf("error\n");
	result = clGetKernelInfo(kernel, CL_KERNEL_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_KERNEL_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
}


void test_cclGetKernelWorkGroupInfo(){/*0x37*/
	cl_kernel kernel;
	cl_device_id device;
	cl_ulong local_mem_size, private_mem_size;
	cl_int result;
	size_t param_size, group_size;

	printf("Testing \"cclGetKernelWorkGroupInfo\"\n");
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	my_scanf1("\tdevice=", "%lx", (long *)&device);
	result = clGetKernelWorkGroupInfo(kernel, device, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &group_size, &param_size);
	printf("\t\tCL_KERNEL_WORK_GROUP_SIZE: ");
	(result == CL_SUCCESS) ? printf("%ld\n", group_size) : printf("error\n");
	result = clGetKernelWorkGroupInfo(kernel, device, CL_KERNEL_LOCAL_MEM_SIZE, sizeof(cl_ulong), &local_mem_size, &param_size);
	printf("\t\tCL_KERNEL_LOCAL_MEM_SIZE: ");
	(result == CL_SUCCESS) ? printf("%ld\n", local_mem_size) : printf("error\n");
	result = clGetKernelWorkGroupInfo(kernel, device, CL_KERNEL_PRIVATE_MEM_SIZE, sizeof(cl_ulong), &private_mem_size, &param_size);
	printf("\t\tCL_KERNEL_PRIVATE_MEM_SIZE: ");
	(result == CL_SUCCESS) ? printf("%ld\n", private_mem_size) : printf("error\n");
}


void test_cclEnqueueNDRangeKernel(){/*0x38*/
	#define MAX_GLOBAL_WORK_OFFSET_ITENS 64
	#define MAX_GLOBAL_WORK_SIZE_ITENS 64
	#define MAX_LOCAL_WORK_SIZE_ITENS 64
	cl_kernel kernel;
	cl_command_queue command_queue;
	cl_int result;
	size_t work_dim, global_work_offset[MAX_GLOBAL_WORK_OFFSET_ITENS], *_global_work_offset, global_work_size[MAX_GLOBAL_WORK_SIZE_ITENS], local_work_size[MAX_LOCAL_WORK_SIZE_ITENS], *_local_work_size;
	char global_work_offsets, local_work_sizes;
	int i;

	printf("Testing \"cclEnqueueNDRangeKernel\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	my_scanf1("\tkernel=", "%lx", (long *)&kernel);
	my_scanf1("\twork_dim=", "%ld", (long *)&work_dim);
	my_scanf1("\tglobal_work_offsets=", "%c", &global_work_offsets);
	my_scanf1("\tlocal_work_sizes=", "%c", &local_work_sizes);
	for(i=0; i<work_dim; i++){
		if(global_work_offsets != 'n')
			my_scanf2("\t\tglobal_work_offset[%d]=", i, "%ld", (long *)&global_work_offset[i]);
		my_scanf2("\t\tglobal_work_size[%d]=", i, "%ld", (long *)&global_work_size[i]);
		if(local_work_sizes != 'n')
			my_scanf2("\t\tlocal_work_size[%d]=", i, "%ld", (long *)&local_work_size[i]);
	}
	_global_work_offset = (global_work_offsets == 'n') ? NULL : global_work_offset;
	_local_work_size = (local_work_sizes == 'n') ? NULL : local_work_size;
	result = clEnqueueNDRangeKernel(command_queue, kernel, work_dim, _global_work_offset, global_work_size, _local_work_size, 0, NULL, NULL);
	printf("\tenqueue: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclCreateUserEvent(){/*0x3B*/
	cl_event event;
	cl_context ctx;
	cl_int errcode_ret;

	printf("Testing \"cclCreateUserEvent\"\n");
	my_scanf1("\tcontext=", "%lx", (long *)&ctx);
	event = clCreateUserEvent(ctx, &errcode_ret);
	printf("\tevent: ");
	(errcode_ret == CL_SUCCESS) ? printf("%#04lx\n", (long)event) : printf("error!\n");
}


void test_cclSetUserEventStatus(){/*0x3C*/
	cl_event event;
	cl_int status, result;

	printf("Testing \"cclSetUserEventStatus\"\n");
	my_scanf1("\tevent=", "%lx", (long *)&event);
	printf("\t%#04x:CL_COMPLETE; %#04x:NEGATIVE_VALUE\n", CL_COMPLETE, -1);
	my_scanf1("\tstatus=", "%x", &status);
	result = clSetUserEventStatus(event, status);
	printf("\tset: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclWaitForEvents(){/*0x3D*/
	#define MAX_EVENTS 8
	cl_event event_list[MAX_EVENTS];
	cl_int result;
	cl_uint num_events;
	int e;

	printf("Testing \"cclWaitForEvents\"\n");
	my_scanf1("\tnum_events=", "%d", &num_events);
	for(e=0; e<num_events; e++)
		my_scanf2("\t\tevent[%d]=", e, "%lx", (long *)&event_list[e]);
	result = clWaitForEvents(num_events, event_list);
	printf("\twait: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclGetEventInfo(){/*0x3E*/
	cl_event event;
	cl_uint result, reference_count;
	cl_context ctx;
	cl_command_queue command_queue;
	cl_command_type command_type;
	cl_int status;
	size_t param_size;

	printf("Testing \"cclGetEventInfo\"\n");
	my_scanf1("\tevent=", "%lx", (long *)&event);
	result = clGetEventInfo(event, CL_EVENT_COMMAND_QUEUE, sizeof(cl_command_queue), &command_queue, &param_size);
	printf("\t\tCL_EVENT_COMMAND_QUEUE: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)command_queue) : printf("error\n");
	result = clGetEventInfo(event, CL_EVENT_CONTEXT, sizeof(cl_context), &ctx, &param_size);
	printf("\t\tCL_EVENT_CONTEXT: ");
	(result == CL_SUCCESS) ? printf("%#04lx\n", (long)ctx) : printf("error\n");
	result = clGetEventInfo(event, CL_EVENT_COMMAND_TYPE, sizeof(cl_command_type), &command_type, &param_size);
	printf("\t\tCL_EVENT_COMMAND_TYPE: ");
	(result == CL_SUCCESS) ? printf("%lx\n", (long)command_type) : printf("error\n");
	result = clGetEventInfo(event, CL_EVENT_COMMAND_EXECUTION_STATUS, sizeof(cl_int), &status, &param_size);
	printf("\t\tCL_EVENT_COMMAND_EXECUTION_STATUS: ");
	(result == CL_SUCCESS) ? printf("%lx\n", (long)status) : printf("error\n");
	result = clGetEventInfo(event, CL_EVENT_REFERENCE_COUNT, sizeof(cl_uint), &reference_count, &param_size);
	printf("\t\tCL_EVENT_REFERENCE_COUNT: ");
	(result == CL_SUCCESS) ? printf("%d\n", reference_count) : printf("error\n");
}


void test_cclRetainEvent(){/*0x40*/
	cl_event event;
	cl_int result;

	printf("Testing \"cclRetainEvent\"\n");
	my_scanf1("\tevent=", "%lx", (long *)&event);
	result = clRetainEvent(event);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclReleaseEvent(){/*0x41*/
	cl_event event;
	cl_int result;

	printf("Testing \"cclReleaseEvent\"\n");
	my_scanf1("\tevent=", "%lx", (long *)&event);
	result = clReleaseEvent(event);
	printf("\tretain: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclFlush(){/*0x46*/
	cl_command_queue command_queue;
	cl_int result;

	printf("Testing \"cclFlush\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	result = clFlush(command_queue);
	printf("\tflush: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


void test_cclFinish(){/*0x47*/
	cl_command_queue command_queue;
	cl_int result;

	printf("Testing \"cclFinish\"\n");
	my_scanf1("\tcommand_queue=", "%lx", (long *)&command_queue);
	result = clFinish(command_queue);
	printf("\tfinish: ");
	(result == CL_SUCCESS) ? printf("ok\n") : printf("error!\n");
}


int main (int argc, char *argv[]){
	void (*test[])() = {/*0x00*/NULL, /*0x01*/test_cclGetPlatformIDs, /*0x02*/test_cclGetPlatformInfo, /*0x03*/test_cclGetDeviceIDs, /*0x04*/test_cclGetDeviceInfo, /*0x05*/test_cclCreateContext, /*0x06*/test_cclCreateContextFromType, /*0x07*/test_cclRetainContext, /*0x08*/test_cclReleaseContext, /*0x09*/test_cclGetContextInfo, /*0x0A*/test_cclCreateCommandQueue, /*0x0B*/test_cclRetainCommandQueue, /*0x0C*/test_cclReleaseCommandQueue, /*0x0D*/test_cclGetCommandQueueInfo, /*0x0E*/test_cclCreateBuffer, /*0x0F*/test_cclCreateSubBuffer, /*0x10*/test_cclEnqueueReadBuffer, /*0x11*/test_cclEnqueueWriteBuffer, /*0x12*/NULL, /*0x13*/NULL, /*0x14*/NULL, /*0x15*/NULL, /*0x16*/NULL, /*0x17*/NULL, /*0x18*/NULL, /*0x19*/NULL, /*0x1A*/NULL, /*0x1B*/NULL, /*0x1C*/NULL, /*0x1D*/NULL, /*0x1E*/NULL, /*0x1F*/NULL, /*0x20*/NULL, /*0x21*/test_cclRetainMemObject, /*0x22*/test_cclReleaseMemObject, /*0x23*/NULL, /*0x24*/NULL, /*0x25*/test_cclGetMemObjectInfo, /*0x26*/NULL, /*0x27*/NULL, /*0x28*/NULL, /*0x29*/test_cclCreateProgramWithSource, /*0x2A*/NULL, /*0x2B*/test_cclRetainProgram, /*0x2C*/test_cclReleaseProgram, /*0x2D*/test_cclBuildProgram, /*0x2E*/NULL, /*0x2F*/test_cclGetProgramInfo, /*0x30*/test_cclGetProgramBuildInfo, /*0x31*/test_cclCreateKernel, /*0x32*/test_cclCreateKernelsInProgram, /*0x33*/test_cclRetainKernel, /*0x34*/test_cclReleaseKernel, /*0x35*/test_cclSetKernelArg, /*0x36*/test_cclGetKernelInfo, /*0x37*/test_cclGetKernelWorkGroupInfo, /*0x38*/test_cclEnqueueNDRangeKernel, /*0x39*/NULL, /*0x3A*/NULL, /*0x3B*/test_cclCreateUserEvent, /*0x3C*/test_cclSetUserEventStatus, /*0x3D*/test_cclWaitForEvents, /*0x3E*/test_cclGetEventInfo, /*0x3F*/NULL, /*0x40*/test_cclRetainEvent, /*0x41*/test_cclReleaseEvent, /*0x42*/NULL, /*0x43*/NULL, /*0x44*/NULL, /*0x45*/NULL, /*0x46*/test_cclFlush, /*0x47*/test_cclFinish};
	int op;

	do{
		printf("--------------------------------------------------------------------------\n");
		printf("0x00:ReadAnswers; 0x01:GetPlatformIDs; 0x02:GetPlatformInfo; 0x03:GetDeviceIDs; 0x04:GetDeviceInfo; 0x05:CreateContext; 0x06:CreateContextFromType; 0x07:RetainContext; 0x08:ReleaseContext; 0x09:GetContextInfo; 0x0A:CreateCommandQueue; 0x0B:RetainCommandQueue; 0x0C:ReleaseCommandQueue; 0x0D:GetCommandQueueInfo; 0x0E:CreateBuffer; 0x0F:CreateSubBuffer; 0x10:EnqueueReadBuffer; 0x11:EnqueueWriteBuffer; 0x12:NULL; 0x13:NULL; 0x14:NULL; 0x15:NULL; 0x16:NULL; 0x17:NULL; 0x18:NULL; 0x19:NULL; 0x1A:NULL; 0x1B:NULL; 0x1C:NULL; 0x1D:NULL; 0x1E:NULL; 0x1F:NULL; 0x20:NULL; 0x21:RetainMemObject; 0x22:ReleaseMemObject; 0x23:NULL; 0x24:NULL; 0x25:GetMemObjectInfo; 0x26:NULL; 0x27:NULL; 0x28:NULL; 0x29:CreateProgramWithSource; 0x2A:NULL; 0x2B:RetainProgram; 0x2C:ReleaseProgram; 0x2D:BuildProgram; 0x2E:NULL; 0x2F:GetProgramInfo; 0x30:GetProgramBuildInfo; 0x31:CreateKernel; 0x32:CreateKernelsInProgram; 0x33:RetainKernel; 0x34:ReleaseKernel; 0x35:SetKernelArg; 0x36:GetKernelInfo; 0x37:GetKernelWorkGroupInfo; 0x38:EnqueueNDRangeKernel; 0x39:NULL; 0x3A:NULL; 0x3B:CreateUserEvent; 0x3C:SetUserEventStatus; 0x3D:WaitForEvents; 0x3E:GetEventInfo; 0x3F:NULL; 0x40:RetainEvent; 0x41:ReleaseEvent; 0x42:NULL; 0x43:NULL; 0x44:NULL; 0x45:NULL; 0x46:Flush; 0x47:Finish; 0xFF:WriteAnswers\n");
		my_scanf1(">?? ", "%x", &op);
		printf("--------------------------------------------------------------------------\n");
		if(op == 0x00)
			read_answers();
		if(op == 0xFF)
			write_answers();
		if((op >= 0x01) && (op <= 0x47))
			test[op]();
	} while(op != -1);

	exit(0);
}

