#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>

#include <open-mx.h>
#include <CL/cl.h>

#include "ccl_wrap.h"
#include "ccl_common.h"


#define NO_NIC_ID 0xFFFFFFFFFFFFFFFF
#define NO_ENDPOINT_ID 0xFFFFFFFF
#define NO_ADDR not_initialized_addr


omx_endpoint_addr_t not_initialized_addr;
uint32_t subtag=0;
pthread_mutex_t subtag_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t _omx_send=PTHREAD_MUTEX_INITIALIZER;
int daemon_count, num_registered_objects=0, max_registered_objects=0;
struct _daemon_info{
	omx_endpoint_addr_t addr;
	uint64_t nic_id;
	uint32_t endpoint_id;
} *daemon_info;
struct _registered_object{
	void *object;
	omx_endpoint_addr_t addr;
	uint64_t nic_id;
	uint32_t endpoint_id;
} *registered_object=NULL;
pthread_mutex_t register_mutex=PTHREAD_MUTEX_INITIALIZER;


uint32_t _omx_next_subtag(){//need global id to support multiple libOpenCCL processes
	uint32_t st;
	
	pthread_mutex_lock(&register_mutex);
	st = ++subtag;
	pthread_mutex_unlock(&register_mutex);
	
	return(st);
}


void * register_object(void *object, omx_endpoint_addr_t addr, uint64_t nic_id, uint32_t endpoint_id){
	int i;

	pthread_mutex_lock(&register_mutex);

	for(i=0; i<num_registered_objects; i++)
		if((registered_object[i].object == object) && (registered_object[i].nic_id == nic_id) && (registered_object[i].endpoint_id == endpoint_id))
			break;
	if(i == num_registered_objects){
		if(num_registered_objects == max_registered_objects){
			max_registered_objects += 128;
			registered_object = realloc(registered_object, sizeof(struct _registered_object) * max_registered_objects);
		}
		registered_object[i].object = object;
		registered_object[i].addr = addr;
		registered_object[i].nic_id = nic_id;
		registered_object[i].endpoint_id = endpoint_id;
		num_registered_objects++;
	}
	
	pthread_mutex_unlock(&register_mutex);

	return((void *)(long)i);
}


void * lookup_object(void *object, uint64_t nic_id, uint32_t endpoint_id){
	int i;

	for(i=0; i<num_registered_objects; i++)
		if((registered_object[i].object == object) && (registered_object[i].nic_id == nic_id) && (registered_object[i].endpoint_id == endpoint_id))
			break;
	if(i == num_registered_objects)
		return(NULL);
	else
		return((void *)(long)i);
}


cl_int __wrap_clGetPlatformIDs(cl_uint num_entries, cl_platform_id *platforms, cl_uint *num_platforms){
	int p, d, num_returned_platforms, _num_returned_platforms;
	cl_uint _num_platforms;
	uint32_t mysubtag;
	struct{
		char id;
		cl_uint num_entries;
	} ccl_request={.id=0x01, .num_entries=num_entries};
	struct{
		cl_int result;
		cl_uint num_platforms;
		cl_platform_id platform[MAX_NUM_PLATFORMS];
	} ccl_reply;

	if((num_platforms == NULL) && (platforms == NULL))
		return(CL_INVALID_VALUE);

	if((ccl_reply.result = __real_clGetPlatformIDs(num_entries, platforms, &_num_platforms)) != CL_SUCCESS)
		return(ccl_reply.result);
	num_returned_platforms = min(_num_platforms, num_entries);
	if(platforms != NULL){
		for(p=0; p<num_returned_platforms; p++)
			platforms[p] = register_object(platforms[p], NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
		ccl_request.num_entries -= num_returned_platforms;
	}
	if(num_platforms != NULL)
		*num_platforms = _num_platforms;

	for(d=0; d<daemon_count; d++){
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), daemon_info[d].addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result != CL_SUCCESS)
			return(ccl_reply.result);
		_num_returned_platforms = min(ccl_reply.num_platforms, ccl_request.num_entries);
		for(p=0; p<_num_returned_platforms; p++)
			if(platforms != NULL)
				platforms[num_returned_platforms++] = register_object(ccl_reply.platform[p], daemon_info[d].addr, daemon_info[d].nic_id, daemon_info[d].endpoint_id);
		ccl_request.num_entries -= _num_returned_platforms;
		if(num_platforms != NULL)
			*num_platforms += ccl_reply.num_platforms;
	}

	return(CL_SUCCESS);
}


cl_int __wrap_clGetPlatformInfo(cl_platform_id platform, cl_platform_info param_name, size_t param_value_size, void *param_value, size_t *param_value_size_ret){
	uint64_t remote_nic_id=NO_NIC_ID;
	uint32_t remote_endpoint_id=NO_ENDPOINT_ID, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		char id;
		cl_platform_id platform;
		cl_platform_info param_name;
		size_t param_value_size;
	} ccl_request={.id=0x02, .platform=NULL, .param_name=param_name, .param_value_size=param_value_size};
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply;

	if(platform != NULL){
		addr = registered_object[(int)(long)platform].addr;
		remote_nic_id = registered_object[(int)(long)platform].nic_id;
		remote_endpoint_id = registered_object[(int)(long)platform].endpoint_id;
		ccl_request.platform = registered_object[(int)(long)platform].object;
	}

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		if(param_name == CL_PLATFORM_HOSTNAME){
			gethostname(param_value, param_value_size);
			((char *)param_value)[param_value_size-1]= '\0';
			*param_value_size_ret = strlen(param_value) + 1;
		}else	
			ccl_reply.result = __real_clGetPlatformInfo(ccl_request.platform, param_name, param_value_size, param_value, param_value_size_ret);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result == CL_SUCCESS){
			if(param_value_size_ret != NULL)
				*param_value_size_ret = ccl_reply.param_value_size_ret;
			if(param_value != NULL)
				memcpy(param_value, ccl_reply.param_value, ccl_reply.param_value_size_ret);
		}
	}

	return(ccl_reply.result);
}


cl_int __wrap_clGetDeviceIDs(cl_platform_id platform, cl_device_type device_type, cl_uint num_entries, cl_device_id *devices, cl_uint *num_devices){
	int d;
	uint64_t remote_nic_id=NO_NIC_ID;
	uint32_t remote_endpoint_id=NO_ENDPOINT_ID, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		char id;
		cl_platform_id platform;
		cl_device_type device_type;
		cl_uint num_entries;
	} ccl_request={.id=0x03, .platform=NULL, .device_type=device_type, .num_entries=num_entries};
	struct{
		cl_int result;
		cl_uint num_devices;
		cl_device_id device[MAX_NUM_DEVICES];
	} ccl_reply;

	if(((num_entries == 0) && (devices != NULL)) || ((num_devices == NULL) && (devices == NULL)))
		return(CL_INVALID_VALUE);

	if(platform != NULL){
		addr = registered_object[(int)(long)platform].addr;
		remote_nic_id = registered_object[(int)(long)platform].nic_id;
		remote_endpoint_id = registered_object[(int)(long)platform].endpoint_id;
		ccl_request.platform = registered_object[(int)(long)platform].object;
	}

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clGetDeviceIDs(ccl_request.platform, device_type, num_entries, devices, &ccl_reply.num_devices);
		if((ccl_reply.result == CL_SUCCESS) && (devices != NULL))
			for(d=0; d<min(ccl_reply.num_devices, num_entries); d++)
				devices[d] = register_object(devices[d], NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result == CL_SUCCESS)
			if(devices != NULL)
				for(d=0; d<min(ccl_reply.num_devices, ccl_request.num_entries); d++)
					devices[d] = register_object(ccl_reply.device[d], addr, remote_nic_id, remote_endpoint_id);
	}
	if(num_devices != NULL)
		*num_devices = ccl_reply.num_devices;

	return(ccl_reply.result);
}


cl_int __wrap_clGetDeviceInfo(cl_device_id device, cl_device_info param_name, size_t param_value_size, void *param_value, size_t *param_value_size_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		int id;
		cl_device_id device;
		cl_device_info param_name;
		size_t param_value_size;
	} ccl_request={.id=0x04, .param_name=param_name, .param_value_size=param_value_size};
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply;

	addr = registered_object[(int)(long)device].addr;
	remote_nic_id = registered_object[(int)(long)device].nic_id;
	remote_endpoint_id = registered_object[(int)(long)device].endpoint_id;
	ccl_request.device = registered_object[(int)(long)device].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID))
		ccl_reply.result = __real_clGetDeviceInfo(ccl_request.device, param_name, param_value_size, param_value, param_value_size_ret);
	else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result == CL_SUCCESS){
			if(param_value_size_ret != NULL)
				*param_value_size_ret = ccl_reply.param_value_size_ret;
			if(param_value != NULL)
				memcpy(param_value, ccl_reply.param_value, ccl_reply.param_value_size_ret);
		}
	}

	return(ccl_reply.result);
}


cl_context __wrap_clCreateContext(const cl_context_properties *properties, cl_uint num_devices, const cl_device_id *devices, void (CL_CALLBACK *pfn_notify)(const char *, const void *, size_t, void *), void *user_data, cl_int *errcode_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int num_properties=0, p, d, num_local_dev=0;
	struct{
		char id;
		cl_context_properties properties[MAX_NUM_PROPERTIES];
		cl_uint num_devices;
		cl_device_id devices[MAX_NUM_DEVICES];
	} ccl_request={.id=0x05, .num_devices=num_devices};
	struct{
		cl_context result;
		cl_int errcode_ret;
	} ccl_reply;

	if(properties != NULL)
		while(properties[num_properties+=2] != 0)
			;
	
	if((num_properties > MAX_NUM_PROPERTIES-1) || (num_devices > MAX_NUM_DEVICES))
		ccl_reply.errcode_ret = CL_OUT_OF_HOST_MEMORY;
	else{
		for(p=0; p<num_properties; p+=2){
			ccl_request.properties[p] = properties[p];
			ccl_request.properties[p+1] = (cl_context_properties)registered_object[(int)(long)properties[p+1]].object;
		}
		ccl_request.properties[p] = 0;
		for(d=0; d<num_devices; d++){
			addr = registered_object[(int)(long)devices[d]].addr;
			remote_nic_id = registered_object[(int)(long)devices[d]].nic_id;
			remote_endpoint_id = registered_object[(int)(long)devices[d]].endpoint_id;
			if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID))
				num_local_dev++;
			ccl_request.devices[d] = registered_object[(int)(long)devices[d]].object;
		}
		
		if(num_local_dev > 0){
			ccl_reply.result = __real_clCreateContext(ccl_request.properties, num_devices, ccl_request.devices, pfn_notify, user_data, &ccl_reply.errcode_ret);
			if(ccl_reply.errcode_ret == CL_SUCCESS)
				ccl_reply.result = register_object(ccl_reply.result, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
		}else{
			mysubtag = _omx_next_subtag();
			_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request) - (MAX_NUM_DEVICES - num_devices) * sizeof(cl_device_id), addr, TAG_REQUEST|mysubtag);
			_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
			if(ccl_reply.errcode_ret == CL_SUCCESS)
				ccl_reply.result = register_object(ccl_reply.result, addr, remote_nic_id, remote_endpoint_id);
		}
	}

	if(errcode_ret != NULL)
		*errcode_ret = ccl_reply.errcode_ret;
	return(ccl_reply.result);
}


cl_command_queue __wrap_clCreateCommandQueue(cl_context context, cl_device_id device, cl_command_queue_properties properties, cl_int *errcode_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		char id;
		cl_context context;
		cl_device_id device;
		cl_command_queue_properties properties;
	} ccl_request={.id=0x0A, .properties=properties};
	struct{
		cl_command_queue result;
		cl_int errcode_ret;
	} ccl_reply;

	addr = registered_object[(int)(long)context].addr;
	remote_nic_id = registered_object[(int)(long)context].nic_id;
	remote_endpoint_id = registered_object[(int)(long)context].endpoint_id;
	ccl_request.context = registered_object[(int)(long)context].object;
	ccl_request.device = registered_object[(int)(long)device].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clCreateCommandQueue(ccl_request.context, ccl_request.device, properties, &ccl_reply.errcode_ret);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, addr, remote_nic_id, remote_endpoint_id);
	}

	if(errcode_ret != NULL)
		*errcode_ret = ccl_reply.errcode_ret;
	return(ccl_reply.result);
}


cl_mem __wrap_clCreateBuffer(cl_context context, cl_mem_flags flags, size_t size, void *host_ptr, cl_int *errcode_ret){
	// not supported: host_ptr!=NULL
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		char id;
		cl_context context;
		cl_mem_flags flags;
		size_t size;
	} ccl_request={.id=0x0E, .flags=flags, .size=size};
	struct{
		cl_mem result;
		cl_int errcode_ret;
	} ccl_reply;

	addr = registered_object[(int)(long)context].addr;
	remote_nic_id = registered_object[(int)(long)context].nic_id;
	remote_endpoint_id = registered_object[(int)(long)context].endpoint_id;
	ccl_request.context = registered_object[(int)(long)context].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clCreateBuffer(ccl_request.context, flags, size, host_ptr, &ccl_reply.errcode_ret);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, addr, remote_nic_id, remote_endpoint_id);
	}

	if(errcode_ret != NULL)
		*errcode_ret = ccl_reply.errcode_ret;
	return(ccl_reply.result);
}


cl_int __wrap_clEnqueueReadBuffer(cl_command_queue command_queue, cl_mem buffer, cl_bool blocking_read, size_t offset, size_t cb, void *ptr, cl_uint num_events_in_wait_list, const cl_event *event_wait_list, cl_event *event){
// not supported: blocking_write==FALSE
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int e, b,cb2;
	struct{
		char id;
		cl_command_queue command_queue;
		cl_mem buffer;
		size_t offset;
		size_t cb;
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} ccl_request={.id=0x10, .offset=offset, .cb=cb, .num_events_in_wait_list=num_events_in_wait_list, .event_is_null=(event==NULL)};
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;

	addr = registered_object[(int)(long)command_queue].addr;
	remote_nic_id = registered_object[(int)(long)command_queue].nic_id;
	remote_endpoint_id = registered_object[(int)(long)command_queue].endpoint_id;
	ccl_request.command_queue = registered_object[(int)(long)command_queue].object;
	ccl_request.buffer = registered_object[(int)(long)buffer].object;
	if(num_events_in_wait_list > MAX_NUM_EVENTS)
		return(CL_OUT_OF_HOST_MEMORY);
	for(e=0; e<num_events_in_wait_list; e++)
		ccl_request.event_wait_list[e] = registered_object[(int)(long)event_wait_list[e]].object;
	
	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clEnqueueReadBuffer(ccl_request.command_queue, ccl_request.buffer, blocking_read, offset, cb, ptr, num_events_in_wait_list, (num_events_in_wait_list == 0) ? NULL : ccl_request.event_wait_list, event);
		if((ccl_reply.result == CL_SUCCESS) && (event != NULL))
			*event = register_object(*event, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result == CL_SUCCESS){
			if(event != NULL)
				*event = register_object(ccl_reply.event, addr, remote_nic_id, remote_endpoint_id);
			for(b=0; b<ccl_request.cb; b+=MAX_OMX_PACKET_SIZE){
				cb2 = min(MAX_OMX_PACKET_SIZE, ccl_request.cb - b);
				_ccl_omx_recv_packet((void *)ptr+b, cb2, TAG_DATA_PACKET|mysubtag, ALL_BITS, NULL);
			}
		}
	}

	return(ccl_reply.result);
}


cl_int __wrap_clEnqueueWriteBuffer(cl_command_queue command_queue, cl_mem buffer, cl_bool blocking_write, size_t offset, size_t cb, const void *ptr, cl_uint num_events_in_wait_list, const cl_event *event_wait_list, cl_event *event){
// not supported: blocking_write==FALSE
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int e, b,cb2;
	struct{
		char id;
		cl_command_queue command_queue;
		cl_mem buffer;
		size_t offset;
		size_t cb;
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} ccl_request={.id=0x11, .offset=offset, .cb=cb, .num_events_in_wait_list=num_events_in_wait_list, .event_is_null=(event==NULL)};
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;

	addr = registered_object[(int)(long)command_queue].addr;
	remote_nic_id = registered_object[(int)(long)command_queue].nic_id;
	remote_endpoint_id = registered_object[(int)(long)command_queue].endpoint_id;
	ccl_request.command_queue = registered_object[(int)(long)command_queue].object;
	ccl_request.buffer = registered_object[(int)(long)buffer].object;
	if(num_events_in_wait_list > MAX_NUM_EVENTS)
		return(CL_OUT_OF_HOST_MEMORY);
	for(e=0; e<num_events_in_wait_list; e++)
		ccl_request.event_wait_list[e] = registered_object[(int)(long)event_wait_list[e]].object;
	
	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clEnqueueWriteBuffer(ccl_request.command_queue, ccl_request.buffer, blocking_write, offset, cb, ptr, num_events_in_wait_list, (num_events_in_wait_list == 0) ? NULL : ccl_request.event_wait_list, event);
		if((ccl_reply.result == CL_SUCCESS) && (event != NULL))
			*event = register_object(*event, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		for(b=0; b<cb; b+=MAX_OMX_PACKET_SIZE){
			cb2 = min(MAX_OMX_PACKET_SIZE, cb - b);
			_ccl_omx_send_packet((void *)ptr+b, cb2, addr, TAG_DATA_PACKET|mysubtag);
		}
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if((ccl_reply.result == CL_SUCCESS) && (event != NULL))
			*event = register_object(ccl_reply.event, addr, remote_nic_id, remote_endpoint_id);
	}

	return(ccl_reply.result);
}


cl_program __wrap_clCreateProgramWithSource(cl_context context, cl_uint count, const char **strings, const size_t *lengths, cl_int *errcode_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int s,	total_length=0;
	struct{
		char id;
		cl_context context;
		cl_uint count;
		size_t lengths[MAX_NUM_LENGTHS];
		char strings[MAX_STRINGS_SIZE];
	} ccl_request={.id=0x29, .count=count};
	struct{
		cl_program result;
		cl_int errcode_ret;
	} ccl_reply;
	char *p=ccl_request.strings;

	addr = registered_object[(int)(long)context].addr;
	remote_nic_id = registered_object[(int)(long)context].nic_id;
	remote_endpoint_id = registered_object[(int)(long)context].endpoint_id;
	ccl_request.context = registered_object[(int)(long)context].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clCreateProgramWithSource(ccl_request.context, count, strings, lengths, &ccl_reply.errcode_ret);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		if(count <= MAX_NUM_LENGTHS)
			for(s=0; s<count; s++){
				ccl_request.lengths[s] = (lengths != NULL) ? lengths[s] : strlen(strings[s] + 1);
				total_length += ccl_request.lengths[s];
				if(total_length < MAX_STRINGS_SIZE)
					memcpy(p, strings[s], ccl_request.lengths[s]);
				p += ccl_request.lengths[s];
			}
		if((count > MAX_NUM_LENGTHS) || (total_length > MAX_STRINGS_SIZE))
			ccl_reply.errcode_ret = CL_OUT_OF_HOST_MEMORY;
		else{
			mysubtag = _omx_next_subtag();
			_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request) - (MAX_STRINGS_SIZE - total_length), addr, TAG_REQUEST|mysubtag);
			_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
			if(ccl_reply.errcode_ret == CL_SUCCESS)
				ccl_reply.result = register_object(ccl_reply.result, addr, remote_nic_id, remote_endpoint_id);
		}
	}

	if(errcode_ret != NULL)
		*errcode_ret = ccl_reply.errcode_ret;
	return(ccl_reply.result);
}


cl_int __wrap_clBuildProgram(cl_program program, cl_uint num_devices, const cl_device_id *device_list, const char *options, void (CL_CALLBACK *pfn_notify)(cl_program, void *), void *user_data){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int d;
	struct{
		char id;
		cl_program program;
		cl_uint num_devices;
		cl_device_id device_list[MAX_NUM_DEVICES];
		char options[MAX_OPTIONS_SIZE];
	} ccl_request={.id=0x2D, .num_devices=num_devices};
	struct{
		cl_int result;
	} ccl_reply;

	if(num_devices > MAX_NUM_DEVICES)
		return(CL_OUT_OF_HOST_MEMORY);
	
	addr = registered_object[(int)(long)program].addr;
	remote_nic_id = registered_object[(int)(long)program].nic_id;
	remote_endpoint_id = registered_object[(int)(long)program].endpoint_id;
	ccl_request.program = registered_object[(int)(long)program].object;
	for(d=0; d<num_devices; d++)
		ccl_request.device_list[d] = registered_object[(int)(long)device_list[d]].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID))
		ccl_reply.result = __real_clBuildProgram(ccl_request.program, num_devices, ccl_request.device_list, options, pfn_notify, user_data);
	else{
		if(strlen(options) > MAX_OPTIONS_SIZE-1)
			return(CL_OUT_OF_HOST_MEMORY);
		strcpy(ccl_request.options, options);
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request) - (MAX_OPTIONS_SIZE-1 - strlen(options)), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
	}

	return(ccl_reply.result);
}


cl_kernel __wrap_clCreateKernel(cl_program program, const char *kernel_name, cl_int *errcode_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		char id;
		cl_program program;
		char kernel_name[MAX_KERNEL_NAME_SIZE];
	} ccl_request={.id=0x31};
	struct{
		cl_kernel result;
		cl_int errcode_ret;
	} ccl_reply;

	addr = registered_object[(int)(long)program].addr;
	remote_nic_id = registered_object[(int)(long)program].nic_id;
	remote_endpoint_id = registered_object[(int)(long)program].endpoint_id;
	ccl_request.program = registered_object[(int)(long)program].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clCreateKernel(ccl_request.program, kernel_name, &ccl_reply.errcode_ret);
		if(ccl_reply.errcode_ret == CL_SUCCESS)
			ccl_reply.result = register_object(ccl_reply.result, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else
		if(strlen(kernel_name) > MAX_KERNEL_NAME_SIZE-1)
			ccl_reply.errcode_ret = CL_OUT_OF_HOST_MEMORY;
		else{
			strcpy(ccl_request.kernel_name, kernel_name);
			mysubtag = _omx_next_subtag();
			_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request) - (MAX_KERNEL_NAME_SIZE-1 - strlen(kernel_name)), addr, TAG_REQUEST|mysubtag);
			_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
			if(ccl_reply.errcode_ret == CL_SUCCESS)
				ccl_reply.result = register_object(ccl_reply.result, addr, remote_nic_id, remote_endpoint_id);
		}

	if(errcode_ret != NULL)
		*errcode_ret = ccl_reply.errcode_ret;
	return(ccl_reply.result);
}


cl_int __wrap_clSetKernelArg(cl_kernel kernel, cl_uint arg_index, size_t arg_size, const void *arg_value){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	char arg_is_memobj;
	struct{
		char id;
		cl_kernel kernel;
		cl_uint arg_index;
		size_t arg_size;
		char arg_value[MAX_ARG_VALUE_SIZE];
	} ccl_request={.id=0x35, .arg_index=arg_index, .arg_size=arg_size};
	struct{
		cl_int result;
	} ccl_reply;

	addr = registered_object[(int)(long)kernel].addr;
	remote_nic_id = registered_object[(int)(long)kernel].nic_id;
	remote_endpoint_id = registered_object[(int)(long)kernel].endpoint_id;
	ccl_request.kernel = registered_object[(int)(long)kernel].object;
	if((arg_size == sizeof(cl_mem)) && (arg_value != NULL) && ((int)(long)(*((cl_mem *)arg_value)) > 0) && ((int)(long)(*((cl_mem *)arg_value)) <= num_registered_objects)){
		*((cl_mem *)ccl_request.arg_value) = registered_object[(int)(long)(*((cl_mem *)arg_value))].object;
		arg_is_memobj = 1;
	}

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID))
		ccl_reply.result = __real_clSetKernelArg(ccl_request.kernel, arg_index, arg_size, arg_is_memobj ? ccl_request.arg_value : arg_value);
	else{
		if(arg_size > MAX_ARG_VALUE_SIZE)
			return(CL_OUT_OF_HOST_MEMORY);
		if(!arg_is_memobj)
			memcpy(ccl_request.arg_value, arg_value, arg_size);
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request) - (MAX_ARG_VALUE_SIZE - arg_size), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
	}

	return(ccl_reply.result);
}


cl_int __wrap_clEnqueueNDRangeKernel(cl_command_queue command_queue, cl_kernel kernel, cl_uint work_dim, const size_t *global_work_offset, const size_t *global_work_size, const size_t *local_work_size, cl_uint num_events_in_wait_list, const cl_event *event_wait_list, cl_event *event){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	int e;
	struct{
		char id;
		cl_command_queue command_queue;
		cl_kernel kernel;
		cl_uint work_dim;
		char global_work_offset_is_null;
		size_t global_work_offset[MAX_NUM_WORK_DIMS];
		size_t global_work_size[MAX_NUM_WORK_DIMS];
		char local_work_size_is_null;
		size_t local_work_size[MAX_NUM_WORK_DIMS];
		cl_uint num_events_in_wait_list;
		cl_event event_wait_list[MAX_NUM_EVENTS];
		char event_is_null;
	} ccl_request={.id=0x38, .work_dim=work_dim, .global_work_offset_is_null=(global_work_offset==NULL), .local_work_size_is_null=(local_work_size==NULL), .num_events_in_wait_list=num_events_in_wait_list, .event_is_null=(event==NULL)};
	struct{
		cl_int result;
		cl_event event;
	} ccl_reply;

	addr = registered_object[(int)(long)command_queue].addr;
	remote_nic_id = registered_object[(int)(long)command_queue].nic_id;
	remote_endpoint_id = registered_object[(int)(long)command_queue].endpoint_id;
	ccl_request.command_queue = registered_object[(int)(long)command_queue].object;
	ccl_request.kernel = registered_object[(int)(long)kernel].object;
	if(num_events_in_wait_list > MAX_NUM_EVENTS)
		return(CL_OUT_OF_HOST_MEMORY);
	for(e=0; e<num_events_in_wait_list; e++)
		ccl_request.event_wait_list[e] = registered_object[(int)(long)event_wait_list[e]].object;
	
	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID)){
		ccl_reply.result = __real_clEnqueueNDRangeKernel(ccl_request.command_queue, ccl_request.kernel, work_dim, global_work_offset, global_work_size, local_work_size, num_events_in_wait_list, (num_events_in_wait_list == 0) ? NULL : ccl_request.event_wait_list, event);
		if((ccl_reply.result == CL_SUCCESS) && (event != NULL))
			*event = register_object(*event, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);
	}else{
		if(work_dim > MAX_NUM_WORK_DIMS)
			return(CL_OUT_OF_HOST_MEMORY);
		if(global_work_offset != NULL)
			memcpy(ccl_request.global_work_offset, global_work_offset, sizeof(size_t) * work_dim);
		memcpy(ccl_request.global_work_size, global_work_size, sizeof(size_t) * work_dim);
		if(local_work_size != NULL)
			memcpy(ccl_request.local_work_size, local_work_size, sizeof(size_t) * work_dim);
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if((ccl_reply.result == CL_SUCCESS) && (event != NULL))
			*event = register_object(ccl_reply.event, addr, remote_nic_id, remote_endpoint_id);
	}

	return(ccl_reply.result);
}


cl_int __wrap_clGetEventProfilingInfo(cl_event event, cl_profiling_info param_name, size_t param_value_size, void *param_value, size_t *param_value_size_ret){
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t addr;
	struct{
		int id;
		cl_event event;
		cl_profiling_info param_name;
		size_t param_value_size;
	} ccl_request={.id=0x48, .param_name=param_name, .param_value_size=param_value_size};
	struct{
		cl_int result;
		size_t param_value_size_ret;
		char param_value[MAX_PARAM_VALUE_SIZE];
	} ccl_reply;

	addr = registered_object[(int)(long)event].addr;
	remote_nic_id = registered_object[(int)(long)event].nic_id;
	remote_endpoint_id = registered_object[(int)(long)event].endpoint_id;
	ccl_request.event = registered_object[(int)(long)event].object;

	if((remote_nic_id == NO_NIC_ID) && (remote_endpoint_id == NO_ENDPOINT_ID))
		ccl_reply.result = __real_clGetEventProfilingInfo(ccl_request.event, param_name, param_value_size, param_value, param_value_size_ret);
	else{
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(&ccl_reply, sizeof(ccl_reply), TAG_REPLY|mysubtag, ALL_BITS, NULL);
		if(ccl_reply.result == CL_SUCCESS){
			if(param_value_size_ret != NULL)
				*param_value_size_ret = ccl_reply.param_value_size_ret;
			if(param_value != NULL)
				memcpy(param_value, ccl_reply.param_value, ccl_reply.param_value_size_ret);
		}
	}

	return(ccl_reply.result);
}


int connect_to_daemons(){
	FILE *fp;
	int count=0;
	uint64_t remote_nic_id;
	uint32_t remote_endpoint_id, mysubtag;
	omx_endpoint_addr_t remote_endpoint_addr;
	struct{
		char id;
	} ccl_request={.id=0x00};

	fp = popen("ccl_daemon_list.py", "r");
	while(fscanf(fp, "%lx; %d\n", &remote_nic_id, &remote_endpoint_id) > 0){
		daemon_info = realloc(daemon_info, sizeof(struct _daemon_info) * ++count);
		if(omx_connect(_ccl_local_endpoint, remote_nic_id, remote_endpoint_id, FILTER_OPENCCL, TIMEOUT_CONNECT, &remote_endpoint_addr))
			_ccl_perror_and_exit("omx_connect");
		daemon_info[count-1].addr = remote_endpoint_addr;
		daemon_info[count-1].nic_id = remote_nic_id;
		daemon_info[count-1].endpoint_id = remote_endpoint_id;
		mysubtag = _omx_next_subtag();
		_ccl_omx_send_packet(&ccl_request, sizeof(ccl_request), daemon_info[count-1].addr, TAG_REQUEST|mysubtag);
		_ccl_omx_recv_packet(NULL, 0, TAG_REPLY|mysubtag, ALL_BITS, NULL);
	}
	pclose(fp);

	return(count);
}


int __wrap_main(int argc, char *argv[]){
	cl_int ret_code;
	cl_platform_id platform;
	cl_uint platforms;

	ret_code = __real_clGetPlatformIDs(1, &platform, &platforms);
	if((ret_code != CL_SUCCESS) || (platforms == 0))
		_ccl_perror_and_exit("OpenCL");

	if(omx_init() != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_init");
	if(omx_open_endpoint(OMX_ANY_NIC, OMX_ANY_ENDPOINT, FILTER_OPENCCL, NULL, 0, &_ccl_local_endpoint) != OMX_SUCCESS)
		_ccl_perror_and_exit("omx_open_endpoint");

	daemon_count = connect_to_daemons();

	register_object(NULL, NO_ADDR, NO_NIC_ID, NO_ENDPOINT_ID);

	exit(__real_main(argc, argv));
}
